import math
from datetime import date

import numpy as np
import pandas as pd
import xarray as xr
from PyQt5.QtWidgets import QFileDialog

from sigmt.__version__ import __version__


def save_edi(estimates: xr.Dataset, procinfo: dict, project_setup: dict) -> None:
    """
    Saves EDI after processing is done.

    :param estimates: Dataset containing estimates
    :type estimates: xarray.Dataset
    :param procinfo: Dictionary containing processing information.
    :type procinfo: dict
    :param project_setup: Dictionary containing project setup.
    :type project_setup: dict

    """

    lat = procinfo['lat']
    if lat > 0:
        lat_d = np.floor(lat)
        lat_m2 = (lat - lat_d) * 60
        lat_m = np.floor(lat_m2)
        lat_s = lat_m2 - lat_m
        del lat_m2
        lat_s = lat_s * 60
    else:
        lat = -1 * lat
        lat_d = np.floor(lat)
        lat_m2 = (lat - lat_d) * 60
        lat_m = np.floor(lat_m2)
        lat_s = lat_m2 - lat_m
        del lat_m2
        lat_s = lat_s * 60
        lat_s = -1 * lat_s
    lon = procinfo['lon']
    if lon > 0:
        lon_d = np.floor(lon)
        lon_m2 = (lon - lon_d) * 60
        lon_m = np.floor(lon_m2)
        lon_s = lon_m2 - lon_m
        del lon_m2
        lon_s = lon_s * 60
    else:
        lon = -1 * lon
        lon_d = np.floor(lon)
        lon_m2 = (lon - lon_d) * 60
        lon_m = np.floor(lon_m2)
        lon_s = lon_m2 - lon_m
        del lon_m2
        lon_s = lon_s * 60
        lon_d = -1 * lon_d

    # Data
    freq = estimates.frequency.values
    # For MT Only and MT + Tipper cases
    if project_setup['processing_mode'] == 'MT + Tipper' or project_setup['processing_mode'] == 'MT Only':
        zxx_r = np.real(estimates['zxx'].values)
        zxx_i = np.imag(estimates['zxx'].values)
        zxy_r = np.real(estimates['zxy'].values)
        zxy_i = np.imag(estimates['zxy'].values)
        zyx_r = np.real(estimates['zyx'].values)
        zyx_i = np.imag(estimates['zyx'].values)
        zyy_r = np.real(estimates['zyy'].values)
        zyy_i = np.imag(estimates['zyy'].values)
        #
        zxx_var = estimates['zxx_var'].values
        zxy_var = estimates['zxy_var'].values
        zyx_var = estimates['zyx_var'].values
        zyy_var = estimates['zyy_var'].values
        #
        coh_ex = estimates['coh_ex'].values
        coh_ey = estimates['coh_ey'].values

    # For Tipper only and MT + Tipper cases
    if project_setup['processing_mode'] == 'MT + Tipper' or project_setup['processing_mode'] == 'Tipper only':
        tx_r = np.real(estimates['tzx'].values)
        tx_i = np.imag(estimates['tzx'].values)
        ty_r = np.real(estimates['tzy'].values)
        ty_i = np.imag(estimates['tzy'].values)
        tx_var = estimates['tzx_var'].values
        ty_var = estimates['tzy_var'].values

    today = date.today()
    today = today.strftime("%m/%d/%Y")
    #
    file_formats = "EDI Files (*.edi);;All Files (*)"
    edi_file, _ = QFileDialog.getSaveFileName(None, "Save File", "", file_formats)
    if edi_file:  # If user selects a path
        with open(edi_file, 'w', encoding='utf-8') as file_handle:
            file_handle.write(">HEAD\n")
            file_handle.write("  DATAID=" + "\"" + procinfo['localsite'] + "\"")
            file_handle.write("\n  ACQBY=" + "\"" + project_setup['acquired_by'] + "\"")
            file_handle.write("\n  FILEBY=" + "\"SigMT\"")
            file_handle.write("\n  ACQDATE=" + pd.to_datetime(procinfo['start_time'], unit='s').strftime("%m/%d/%Y"))
            file_handle.write("\n  FILEDATE=" + today)
            file_handle.write(f"\n  PROSPECT=\"{project_setup['project_name']}\"")
            file_handle.write(
                "\n  LAT=" + str(int(lat_d)) + ":" + str(int(np.floor(lat_m))) + ":" + str(round(lat_s, 2)))
            file_handle.write(
                "\n  LONG=" + str(int(lon_d)) + ":" + str(int(np.floor(lon_m))) + ":" + str(round(lon_s, 2)))
            file_handle.write("\n  ELEV=" + str(procinfo['elev']))
            file_handle.write("\n  STDVERS=\"SEG 1.0\"")
            file_handle.write("\n  PROGVERS=\"" + __version__ + "\"")
            file_handle.write("\n  PROGDATE= ")
            file_handle.write("\n  MAXSECT=999")
            file_handle.write("\n  EMPTY=1.000000000E+32")

            # ===== INFO section =====
            file_handle.write("\n\n>INFO\n")
            file_handle.write("  MAXINFO=500")
            file_handle.write("\n\n")
            file_handle.write("------------------------------------------------------------")

            # ===== DEFINEMEAS section =====
            file_handle.write("\n\n>=DEFINEMEAS\n")
            file_handle.write("  REFLOC=" + "\"" + procinfo['localsite'] + "\"")
            file_handle.write("\n  REFLAT=" + str(round(lat_d)) + ":" + str(round(lat_m)) + ":" + str(round(lat_s, 2)))
            file_handle.write("\n  REFLONG=" + str(round(lon_d)) + ":" + str(round(lon_m)) + ":" + str(round(lon_s, 2)))
            file_handle.write("\n  REFELEV=" + str(procinfo['elev']))

            if project_setup['processing_mode'] == 'MT + Tipper':
                file_handle.write(
                    "\n"
                    "\n>EMEAS ID=" + "101.001" + " CHTYPE=EX X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                 "Y2=0.000000e+00 Z2=0.000000e+00")
                file_handle.write(
                    "\n>EMEAS ID=" + "102.001" + " CHTYPE=EY X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                 "Y2=0.000000e+00 Z2=0.000000e+00")
                file_handle.write(
                    "\n>HMEAS ID=" + "103.001" + " CHTYPE=HX X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                 "Y2=0.000000e+00 Z2=0.000000e+00")
                file_handle.write(
                    "\n>HMEAS ID=" + "104.001" + " CHTYPE=HY X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                 "Y2=0.000000e+00 Z2=0.000000e+00")
                file_handle.write(
                    "\n>HMEAS ID=" + "105.001" + " CHTYPE=HZ X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                 "Y2=0.000000e+00 Z2=0.000000e+00")
                if procinfo['remotesite'] is not None:
                    file_handle.write(
                        "\n>HMEAS ID=" + "106.001" + " CHTYPE=HX X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                     "Y2=0.000000e+00 Z2=0.000000e+00")
                    file_handle.write(
                        "\n>HMEAS ID=" + "107.001" + " CHTYPE=HY X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                     "Y2=0.000000e+00 Z2=0.000000e+00")
            elif project_setup['processing_mode'] == 'MT only':
                file_handle.write(
                    "\n"
                    "\n>EMEAS ID=" + "101.001" + " CHTYPE=EX X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                 "Y2=0.000000e+00 Z2=0.000000e+00")
                file_handle.write(
                    "\n>EMEAS ID=" + "102.001" + " CHTYPE=EY X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                 "Y2=0.000000e+00 Z2=0.000000e+00")
                file_handle.write(
                    "\n>HMEAS ID=" + "103.001" + " CHTYPE=HX X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                 "Y2=0.000000e+00 Z2=0.000000e+00")
                file_handle.write(
                    "\n>HMEAS ID=" + "104.001" + " CHTYPE=HY X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                 "Y2=0.000000e+00 Z2=0.000000e+00")
                if procinfo['remotesite'] is not None:
                    file_handle.write(
                        "\n>HMEAS ID=" + "106.001" + " CHTYPE=HX X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                     "Y2=0.000000e+00 Z2=0.000000e+00")
                    file_handle.write(
                        "\n>HMEAS ID=" + "107.001" + " CHTYPE=HY X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                     "Y2=0.000000e+00 Z2=0.000000e+00")
            elif project_setup['processing_mode'] == 'Tipper only':
                file_handle.write(
                    "\n>HMEAS ID=" + "103.001" + " CHTYPE=HX X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                 "Y2=0.000000e+00 Z2=0.000000e+00")
                file_handle.write(
                    "\n>HMEAS ID=" + "104.001" + " CHTYPE=HY X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                 "Y2=0.000000e+00 Z2=0.000000e+00")
                if procinfo['remotesite'] is not None:
                    file_handle.write(
                        "\n>HMEAS ID=" + "106.001" + " CHTYPE=HX X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                     "Y2=0.000000e+00 Z2=0.000000e+00")
                    file_handle.write(
                        "\n>HMEAS ID=" + "107.001" + " CHTYPE=HY X=0.000000e+00 Y=0.000000e+00 Z=0.000000e+00 X2=0.000000e+00 "
                                                     "Y2=0.000000e+00 Z2=0.000000e+00")

            # ===== MTSECT section =====
            file_handle.write("\n\n>=MTSECT\n")
            file_handle.write("  SECTID=" + "\"" + procinfo['localsite'] + "\"")
            file_handle.write("\n  NFREQ=" + str(len(estimates.frequency)))
            if project_setup['processing_mode'] == 'MT + Tipper':
                file_handle.write("\n  EX=" + "101.001")
                file_handle.write("\n  EY=" + "102.001")
                file_handle.write("\n  HX=" + "103.001")
                file_handle.write("\n  HY=" + "104.001")
                file_handle.write("\n  HZ=" + "105.001")
                if procinfo['remotesite'] is not None:
                    file_handle.write("\n  RX=" + "106.001")
                    file_handle.write("\n  RY=" + "107.001")
            elif project_setup['processing_mode'] == 'MT Only':
                file_handle.write("\n  EX=" + "101.001")
                file_handle.write("\n  EY=" + "102.001")
                file_handle.write("\n  HX=" + "103.001")
                file_handle.write("\n  HY=" + "104.001")
                if procinfo['remotesite'] is not None:
                    file_handle.write("\n  RX=" + "106.001")
                    file_handle.write("\n  RY=" + "107.001")
            elif project_setup['processing_mode'] == 'Tipper only':
                file_handle.write("\n  HX=" + "103.001")
                file_handle.write("\n  HY=" + "104.001")
                file_handle.write("\n  HZ=" + "105.001")
                if procinfo['remotesite'] is not None:
                    file_handle.write("\n  RX=" + "106.001")
                    file_handle.write("\n  RY=" + "107.001")

            # ===== FREQ section =====
            file_handle.write("\n \n")
            file_handle.write(">FREQ  //" + str(np.size(freq)))
            file_handle.write("\n")

            for i in range(np.size(freq)):
                k = i + 1
                if freq[i] < 0:
                    file_handle.write("%.9E " % (freq[i]))
                else:
                    file_handle.write(" %.9E " % (freq[i]))
                if (k % 6 == 0) and (k != 0):
                    file_handle.write("\n")

            if project_setup['processing_mode'] == 'MT + Tipper' or project_setup['processing_mode'] == 'MT Only':
                # ===== ZXXR section =====
                file_handle.write("\n \n")
                file_handle.write(">ZXXR  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if zxx_r[i] < 0:
                        file_handle.write("%.9E " % float(zxx_r[i]))
                    else:
                        file_handle.write(" %.9E " % float(zxx_r[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== ZXXI section =====
                file_handle.write("\n \n")
                file_handle.write(">ZXXI  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if zxx_i[i] < 0:
                        file_handle.write("%.9E " % float(zxx_i[i]))
                    else:
                        file_handle.write(" %.9E " % float(zxx_i[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== ZXX.VAR section =====
                file_handle.write("\n \n")
                file_handle.write(">ZXX.VAR  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if zxx_var[i] < 0:
                        file_handle.write("%.9E " % (zxx_var[i]))
                    else:
                        file_handle.write(" %.9E " % (zxx_var[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== ZXYR section =====
                file_handle.write("\n \n")
                file_handle.write(">ZXYR  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if zxy_r[i] < 0:
                        file_handle.write("%.9E " % float(zxy_r[i]))
                    else:
                        file_handle.write(" %.9E " % float(zxy_r[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== ZXYI section =====
                file_handle.write("\n \n")
                file_handle.write(">ZXYI  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if zxy_i[i] < 0:
                        file_handle.write("%.9E " % float(zxy_i[i]))
                    else:
                        file_handle.write(" %.9E " % float(zxy_i[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== ZXYI.VAR section =====
                file_handle.write("\n \n")
                file_handle.write(">ZXY.VAR  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if zxy_var[i] < 0:
                        file_handle.write("%.9E " % (zxy_var[i]))
                    else:
                        file_handle.write(" %.9E " % (zxy_var[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== ZYXR section =====
                file_handle.write("\n \n")
                file_handle.write(">ZYXR  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if zyx_r[i] < 0:
                        file_handle.write("%.9E " % float(zyx_r[i]))
                    else:
                        file_handle.write(" %.9E " % float(zyx_r[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== ZYXI section =====
                file_handle.write("\n \n")
                file_handle.write(">ZYXI  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if zyx_i[i] < 0:
                        file_handle.write("%.9E " % float(zyx_i[i]))
                    else:
                        file_handle.write(" %.9E " % float(zyx_i[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== ZYX.VAR section =====
                file_handle.write("\n \n")
                file_handle.write(">ZYX.VAR  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if zyx_var[i] < 0:
                        file_handle.write("%.9E " % float(zyx_var[i]))
                    else:
                        file_handle.write(" %.9E " % float(zyx_var[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== ZYYR section =====
                file_handle.write("\n \n")
                file_handle.write(">ZYYR  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if zyy_r[i] < 0:
                        file_handle.write("%.9E " % float(zyy_r[i]))
                    else:
                        file_handle.write(" %.9E " % float(zyy_r[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== ZYYI section =====
                file_handle.write("\n \n")
                file_handle.write(">ZYYI  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if zyy_i[i] < 0:
                        file_handle.write("%.9E " % float(zyy_i[i]))
                    else:
                        file_handle.write(" %.9E " % float(zyy_i[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== ZYY.VAR section =====
                file_handle.write("\n \n")
                file_handle.write(">ZYY.VAR  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if zyy_var[i] < 0:
                        file_handle.write("%.9E " % float(zyy_var[i]))
                    else:
                        file_handle.write(" %.9E " % float(zyy_var[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

            if project_setup['processing_mode'] == 'MT + Tipper' or project_setup['processing_mode'] == 'Tipper Only':
                # ===== TXR.EXP section =====
                file_handle.write("\n \n")
                file_handle.write(">TXR.EXP  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if tx_r[i] < 0:
                        file_handle.write("%.9E " % float(tx_r[i]))
                    else:
                        file_handle.write(" %.9E " % float(tx_r[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== TXI.EXP section =====
                file_handle.write("\n \n")
                file_handle.write(">TXI.EXP  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if tx_i[i] < 0:
                        file_handle.write("%.9E " % float(tx_i[i]))
                    else:
                        file_handle.write(" %.9E " % float(tx_i[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== TXVAR.EXP section =====
                file_handle.write("\n \n")
                file_handle.write(">TXVAR.EXP  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if tx_var[i] < 0:
                        file_handle.write("%.9E " % float(tx_var[i]))
                    else:
                        file_handle.write(" %.9E " % float(tx_var[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== TYR.EXP section =====
                file_handle.write("\n \n")
                file_handle.write(">TYR.EXP  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if ty_r[i] < 0:
                        file_handle.write("%.9E " % float(ty_r[i]))
                    else:
                        file_handle.write(" %.9E " % float(ty_r[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== TYI.EXP section =====
                file_handle.write("\n \n")
                file_handle.write(">TYI.EXP  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if ty_i[i] < 0:
                        file_handle.write("%.9E " % float(ty_i[i]))
                    else:
                        file_handle.write(" %.9E " % float(ty_i[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

                # ===== TYVAR.EXP section =====
                file_handle.write("\n \n")
                file_handle.write(">TYVAR.EXP  //" + str(np.size(freq)))
                file_handle.write("\n")

                for i in range(np.size(freq)):
                    k = i + 1
                    if ty_var[i] < 0:
                        file_handle.write("%.9E " % float(ty_var[i]))
                    else:
                        file_handle.write(" %.9E " % float(ty_var[i]))
                    if (k % 6 == 0) and (k != 0):
                        file_handle.write("\n")

            if project_setup['processing_mode'] == 'MT + Tipper' or project_setup['processing_mode'] == 'MT Only':
                if procinfo['remotesite'] is not None:
                    # ===== cohEx section =====
                    file_handle.write("\n \n")
                    file_handle.write(
                        ">COH  MEAS1=" + "101.001" + " MEAS2=" + "107.001" + "  ROT=NORTH //" + str(np.size(freq)))
                    file_handle.write("\n")

                    for i in range(np.size(freq)):
                        k = i + 1
                        if coh_ex[i] < 0:
                            file_handle.write("%.9E " % float(coh_ex[i]))
                        else:
                            file_handle.write(" %.9E " % float(coh_ex[i]))
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                    # ===== cohEy section =====
                    file_handle.write("\n \n")
                    file_handle.write(
                        ">COH  MEAS1=" + "102.001" + " MEAS2=" + "106.001" + "  ROT=NORTH //" + str(np.size(freq)))
                    file_handle.write("\n")

                    for i in range(np.size(freq)):
                        k = i + 1
                        if coh_ey[i] < 0:
                            file_handle.write("%.9E " % float(coh_ey[i]))
                        else:
                            file_handle.write(" %.9E " % float(coh_ey[i]))
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")
                else:
                    # ===== cohEx section =====
                    file_handle.write("\n \n")
                    file_handle.write(
                        ">COH  MEAS1=" + "101.001" + " MEAS2=" + "105.001" + "  ROT=NORTH //" + str(np.size(freq)))
                    file_handle.write("\n")

                    for i in range(np.size(freq)):
                        k = i + 1
                        if coh_ex[i] < 0:
                            file_handle.write("%.9E " % float(coh_ex[i]))
                        else:
                            file_handle.write(" %.9E " % float(coh_ex[i]))
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

                    # ===== cohEy section =====
                    file_handle.write("\n \n")
                    file_handle.write(
                        ">COH  MEAS1=" + "102.001" + " MEAS2=" + "104.001" + "  ROT=NORTH //" + str(np.size(freq)))
                    file_handle.write("\n")

                    for i in range(np.size(freq)):
                        k = i + 1
                        if coh_ey[i] < 0:
                            file_handle.write("%.9E " % float(coh_ey[i]))
                        else:
                            file_handle.write(" %.9E " % float(coh_ey[i]))
                        if (k % 6 == 0) and (k != 0):
                            file_handle.write("\n")

            file_handle.write("\n \n")
            file_handle.write(">END")
            file_handle.close()


def read_edi(file_path: str) -> tuple:
    """
    Reads EDI file

    :param file_path: EDI file path
    :type file_path: str
    :return: Data from EDI
    :rtype: tuple

    """
    data = {}
    f = open(file_path, 'r')
    edi_cont = f.readlines()

    # # Iterate through each line in the file
    # for line in edi_cont:
    #     line = line.strip()  # Remove leading and trailing whitespaces
    #     if line.startswith('REFLAT'):
    #         reflat = line.split('=')[1]
    #         flat = [float(i) for i in reflat.split(':')]
    #         lat = flat[0] + flat[1] / 60 + flat[2] / (60 * 60)
    #     if line.startswith('REFLONG'):
    #         reflat = line.split('=')[1]
    #         flong = [float(i) for i in reflat.split(':')]
    #         long = flong[0] + flong[1] / 60 + flong[2] / (60 * 60)
    #     if line.startswith('NFREQ'):
    #         nfreq = int(line.split('=')[1])

    # Read Frequencies
    for index, line in enumerate(edi_cont):
        line = line.strip()
        if line.startswith('>FREQ'):
            break
    freqs = []
    for i in range(index + 1, len(edi_cont)):
        line = edi_cont[i]
        try:
            values = [float(val) for val in line.split()]
            freqs.extend(values)
        except ValueError:
            break
    data['freqs'] = freqs

    # Number of coils
    hmeas = 0
    for line in edi_cont:
        if line.startswith('>HMEAS'):
            hmeas += 1

    # =========== MT Impedance =====================
    data['zxx_r'] = read_component('>ZXXR', freqs, edi_cont)
    data['zxx_i'] = read_component('>ZXXI', freqs, edi_cont)
    data['zxx_var'] = read_component('>ZXX.VAR', freqs, edi_cont)
    data['zxy_r'] = read_component('>ZXYR', freqs, edi_cont)
    data['zxy_i'] = read_component('>ZXYI', freqs, edi_cont)
    data['zxy_var'] = read_component('>ZXY.VAR', freqs, edi_cont)
    data['zyx_r'] = read_component('>ZYXR', freqs, edi_cont)
    data['zyx_i'] = read_component('>ZYXI', freqs, edi_cont)
    data['zyx_var'] = read_component('>ZYX.VAR', freqs, edi_cont)
    data['zyy_r'] = read_component('>ZYYR', freqs, edi_cont)
    data['zyy_i'] = read_component('>ZYYI', freqs, edi_cont)
    data['zyy_var'] = read_component('>ZYY.VAR', freqs, edi_cont)

    # =========== Tipper =====================
    data['tzx_r'] = read_component('>TXR.EXP', freqs, edi_cont)
    data['tzx_i'] = read_component('>TXI.EXP', freqs, edi_cont)
    data['tzx_var'] = read_component('>TXVAR.EXP', freqs, edi_cont)
    data['tzy_r'] = read_component('>TYR.EXP', freqs, edi_cont)
    data['tzy_i'] = read_component('>TYI.EXP', freqs, edi_cont)
    data['tzy_var'] = read_component('>TYVAR.EXP', freqs, edi_cont)

    # =========== Coherency =====================
    data['coh_ex'] = read_component('>COH  MEAS1=101.001', freqs, edi_cont)
    data['coh_ey'] = read_component('>COH  MEAS1=102.001', freqs, edi_cont)
    f.close()
    return edi_cont, data, hmeas


def read_component(text: str, freqs: list, edi_cont: list) -> list:
    """
    Read components from EDI file.

    :param text: Component name. Eg: '>ZXXR'
    :type text: str
    :param freqs: List of frequencies
    :type freqs: list
    :param edi_cont: EDI content
    :type edi_cont: list
    :return: Data from the EDI
    :rtype: list

    """
    data = []
    for index, line in enumerate(edi_cont):
        line = line.strip()
        if line.startswith(text):
            break
    for i in range(index + 1, len(edi_cont)):
        line = edi_cont[i]
        try:
            values = [float(val) for val in line.split()]
            data.extend(values)
        except ValueError:
            break
    if not data:
        data = [math.nan * int(freq) for freq in freqs]
    return data



"""
Module for signal processing operations
"""

from typing import Optional

import numpy as np
from scipy import signal


def notch_filter_sos(time_series: np.ndarray,
                     sampling_frequency: float,
                     notch_frequency: float,
                     harmonics: Optional[int] = None) -> np.ndarray:
    """
    SOS filter for notch filtering. Filters out the requested notch frequency and harmonics.

    :param time_series: Time series data
    :type time_series: np.ndarray
    :param sampling_frequency: Sampling frequency
    :type sampling_frequency: float
    :param notch_frequency: Frequency to be filtered
    :type notch_frequency: float
    :param harmonics: Number of harmonics to be filtered out.
    :type harmonics: Optional[int]

    :return: Filtered time series data
    :rtype: np.ndarray

    """
    min_fs = (notch_frequency + 5) * 2
    if sampling_frequency > min_fs:
        if harmonics is None:
            harmonics = int((sampling_frequency / 2.5) / notch_frequency) + 1
        if harmonics > int(14000 / notch_frequency):
            harmonics = int(14000 / notch_frequency)
        print('No. of harmonics: ' + str(harmonics))
        sos = signal.butter(4, [notch_frequency - 5, notch_frequency + 5], btype='bandstop', fs=sampling_frequency,
                            output='sos')
        for n in range(2, harmonics + 1):
            f0 = n * notch_frequency
            sos_new = signal.butter(4, [f0 - 5, f0 + 5], btype='bandstop', fs=sampling_frequency, output='sos')
            sos = np.concatenate((sos, sos_new), axis=0)
        time_series = signal.sosfilt(sos, time_series, axis=0)
    else:
        print("Skipping notch filter as it cannot be performed for this sampling frequency")

    return time_series


def do_fft(ts: np.ndarray, fs: float, fft_length: int) -> tuple:
    """
    Function to perform FFT.

    :param ts: Time series data
    :type ts: np.ndarray
    :param fs: Sampling frequency
    :type fs: float
    :param fft_length: FFT length
    :type fft_length: int
    :return: Tuple of f[np.ndarray: shape n,1], xfft[np.ndarray: shape n,n]
    :rtype: tuple

    """
    w = np.hanning(fft_length).reshape(-1, 1)
    fft_value = np.fft.fft(ts * w, fft_length, axis=0)
    xfft = np.asarray(fft_value[0:int(fft_length / 2), :])
    fline = np.asarray([np.linspace(0, int(fft_length / 2), num=int(fft_length / 2), dtype=int)])
    f = np.asarray(fline * fs / fft_length).T
    f = f[:, 0]

    return f, xfft



import numpy as np
import xarray as xr


def perform_coh_thresh(bandavg_dataset: xr.Dataset, coh_thresh: float, min_percent: float,
                       channels: list) -> xr.Dataset:
    """
    Performs cohrency thresholding

    :param bandavg_dataset: Band averaged dataset
    :type bandavg_dataset: xr.Dataset
    :param coh_thresh: Coherency threshold values
    :type coh_thresh: float
    :param min_percent: Minimum percentage of data need to be maintained
    :type min_percent: float
    :param channels: List of channels in which thresholding to be applied.
    :type channels: list
    :return: Band averaged dataset added with coherency selection xarrays
    :rtype: xr.Dataset

    """
    for channel in channels:
        min_windows = int(
            np.ceil(len(bandavg_dataset[f'coh_{channel}'].coords['time_window']) * (min_percent / 100)))
        for ft in bandavg_dataset[f'coh_{channel}'].coords['frequency']:
            ct = coh_thresh
            coh = bandavg_dataset[f'coh_{channel}'].sel(frequency=ft)
            while np.sum(coh >= ct) < min_windows:
                ct = ct - 0.01
            selection_array = coh > ct
            bandavg_dataset[f'{channel}_selection_coh'].loc[dict(frequency=ft)] = selection_array

    return bandavg_dataset


def clear_coh_thresh(bandavg_dataset: xr.Dataset) -> xr.Dataset:
    """
    Clears coherency threshold

    :param bandavg_dataset: Band averaged dataset
    :type bandavg_dataset: xr.Dataset

    """
    if 'ex_selection_coh' in bandavg_dataset:
        bandavg_dataset['ex_selection_coh'] = xr.DataArray(
            np.full(bandavg_dataset['ex_selection_coh'].shape, True),
            coords=bandavg_dataset.coords,
            dims=bandavg_dataset.dims
        )
    if 'ey_selection_coh' in bandavg_dataset:
        bandavg_dataset['ey_selection_coh'] = xr.DataArray(
            np.full(bandavg_dataset['ey_selection_coh'].shape, True),
            coords=bandavg_dataset.coords,
            dims=bandavg_dataset.dims
        )
    if 'hz_selection_coh' in bandavg_dataset:
        bandavg_dataset['hz_selection_coh'] = xr.DataArray(
            np.full(bandavg_dataset['hz_selection_coh'].shape, True),
            coords=bandavg_dataset.coords,
            dims=bandavg_dataset.dims
        )
    return bandavg_dataset


def perform_pd_selection(bandavg_dataset: xr.Dataset, component: str, pd_min: float, pd_max: float) -> xr.Dataset:
    """
    Performs cohrency thresholding

    :param bandavg_dataset: Band averaged dataset
    :type bandavg_dataset: xr.Dataset
    :param component: Electric or Magnetic component
    :type component: str
    :param pd_min: Minimum value for polarization direction
    :type pd_min: float
    :param pd_max: Maximum value for polarization direction
    :type pd_max: float
    :return: Band averaged dataset added with coherency selection xarrays
    :rtype: xr.Dataset

    """
    if component == 'Electric':
        for ft in bandavg_dataset['alpha_e'].coords['frequency']:
            alpha_e = bandavg_dataset['alpha_e'].sel(frequency=ft)
            selection_array = (alpha_e < pd_min) & (alpha_e > pd_max)
            bandavg_dataset['alpha_e_selection'].loc[dict(frequency=ft)] = selection_array
    elif component == 'Magnetic':
        for ft in bandavg_dataset['alpha_h'].coords['frequency']:
            alpha_h = bandavg_dataset['alpha_h'].sel(frequency=ft)
            selection_array = (alpha_h < pd_min) & (alpha_h > pd_max)
            bandavg_dataset['alpha_h_selection'].loc[dict(frequency=ft)] = selection_array
    return bandavg_dataset


def clear_pd_selection(bandavg_dataset: xr.Dataset) -> xr.Dataset:
    """
    Clears Polarization direction based selection

    :param bandavg_dataset: Band averaged dataset
    :type bandavg_dataset: xr.Dataset

    """
    if 'alpha_e_selection' in bandavg_dataset:
        bandavg_dataset['alpha_e_selection'] = xr.DataArray(
            np.full(bandavg_dataset['ex'].shape, True),
            coords=bandavg_dataset.coords,
            dims=bandavg_dataset.dims
        )

    if 'alpha_h_selection' in bandavg_dataset:
        bandavg_dataset['alpha_h_selection'] = xr.DataArray(
            np.full(bandavg_dataset['hx'].shape, True),
            coords=bandavg_dataset.coords,
            dims=bandavg_dataset.dims
        )
    return bandavg_dataset



"""
Class for band averaging. Here, auto and cross spectra are calculated for
each target frequencies and transfer functions are computed for all events.
"""

import gc
from concurrent.futures import ThreadPoolExecutor
from typing import Optional

import numpy as np
import xarray as xr
from scipy import signal

import sigmt.core.sigproc as sp
import sigmt.core.statistics as stats
import sigmt.utils.utils as utils
from sigmt.utils.metronix.calibration import MetronixCalibration


class BandAveraging:
    """
    Class to perform band averaging.

    """

    def __init__(self,
                 time_series: dict[str, np.ndarray],
                 sampling_frequency: float,
                 fft_length: int = 1024,
                 parzen_window_radius: float = 0.25,
                 overlap: int = 50,
                 frequencies_per_decade: int = 12,
                 remote_reference: bool = False,
                 calibrate_electric: bool = False,
                 calibrate_magnetic: bool = False,
                 calibration_data_electric: Optional[dict[str, dict[str, float]]] = None,
                 calibration_data_magnetic: Optional[dict] = None,
                 apply_notch_filter: bool = False,
                 notch_frequency: Optional[float] = None,
                 notch_harmonics: Optional[int] = None,
                 process_mt: bool = True,
                 process_tipper: bool = True) -> None:
        """
        Constructor

        This class performs band averaging. It divides the time series into different time
        windows or segments, applies notch filtering, calibration (if required), averages
        the cross powers for the target frequencies using a parzen window.

        TODO: Improve error handling

        :param time_series: A dictionary containing time series data as 1D numpy arrays.
                    Each key represents a component of the data, and the corresponding
                    value is a numpy array of time series values. The dictionary should
                    include the following keys and associated units:
                    - 'ex': Electric field component in the x-direction
                        - Unit: mV/km (if calibrated), mV (Metronix, if not calibrated)
                    - 'ey': Electric field component in the y-direction
                        - Unit: mV/km (if calibrated), mV (Metronix, if not calibrated)
                    - 'hx': Magnetic field component in the x-direction
                        - Unit: nT (if calibrated), mV (Metronix, if not calibrated)
                    - 'hy': Magnetic field component in the y-direction
                        - Unit: nT (if calibrated), mV (Metronix, if not calibrated)
                    - 'hz': Magnetic field component in the z-direction
                        - Unit: nT (if calibrated), mV (Metronix, if not calibrated)
                    - 'rx': Remote magnetic field component in the x-direction
                        - Unit: nT (if calibrated), mV (Metronix, if not calibrated)
                    - 'ry': Remote magnetic field component in the y-direction
                        - Unit: nT (if calibrated), mV (Metronix, if not calibrated)

        :type time_series: dict[str, numpy.ndarray]
        :param sampling_frequency: Sampling frequency of the time_series.
        :type sampling_frequency: float
        :param fft_length: The length of the Fast Fourier Transform (FFT) to be applied.
                           The default value is 1024.
        :type fft_length: int
        :param parzen_window_radius: The radius of the Parzen window used for band averaging.
                                     The default value is 0.25.
        :type parzen_window_radius: float
        :param overlap: The percentage of overlap between consecutive time windows when segmenting the time series.
                        The default value is 50%.
        :type overlap: int
        :param frequencies_per_decade: The number of target frequencies to be used per decade for band averaging.
                                       The default value is 12.
        :type frequencies_per_decade: int
        :param remote_reference: A boolean flag indicating whether remote reference is used for processing.
                                 Set to True if a remote reference is used. The default is False.
        :type remote_reference: bool
        :param calibrate_electric: A boolean flag that specifies whether the electric channels should be calibrated
                                   from units of mV to mv/km. This needs dipole length details. The default is False.
        :type calibrate_electric: bool
        :param calibrate_magnetic: A boolean flag that specifies whether the magnetic channels should be calibrated
                                   to units of mV to nT. The calibration process currently supports Metronix sensor
                                   coils. The default is False.
        :type calibrate_magnetic: bool
        :param calibration_data_electric: A dictionary containing the distances from the center to each electrode.
                                          It should include two sub-dictionaries with the following distances:
                                          - 'ex':
                                              - 'x1': Distance (float) from the center to the North electrode
                                              - 'x2': Distance (float) from the center to the South electrode
                                          - 'ey':
                                              - 'y1': Distance (float) from the center to the East electrode
                                              - 'y2': Distance (float) from the center to the West electrode
        :type calibration_data_electric: dict[str, dict[str, float]]
        :param calibration_data_magnetic: A dictionary containing all information about the coil details.
                                          Currently supports only Metronix.
                                          # TODO: Add more instrument support
                                          Mandatory keys:
                                          - 'instrument': 'metronix' (str)
                                          - 'hx' (dict) details of hx coil
                                                - 'sensor_type': 'MFS06e' (str)
                                                - 'sensor_serial_number': 300 (int)
                                                - 'chopper_status': 1 (int) 1 means 0N, 0 means OFF
                                                - 'calibration_data' (dict)
                                                    - '300' (dict) - key is coil number
                                                        - 'chopper_on' (ndarray, shape: n,3)
                                                           n - number of frequencies
                                                           First column: Frequency (Hz)
                                                           Second column: Magnitude (V/nT*Hz)
                                                           Third column: Phase  (deg)
                                                        - 'chopper_off' (ndarray, shape: n,3)
                                                           n - number of frequencies
                                                           First column: Frequency (Hz)
                                                           Second column: Magnitude (V/nT*Hz)
                                                           Third column: Phase  (deg)
                                          - 'hy' (dict) details of hx coil
                                                - 'sensor_type': 'MFS06e' (str)
                                                - 'sensor_serial_number': 301 (int)
                                                - 'chopper_status': 1 (int) 1 means 0N, 0 means OFF
                                                - 'calibration_data' (dict)
                                                    - '301' (dict) - key is coil number
                                                        - 'chopper_on' (ndarray, shape: n,3)
                                                           n - number of frequencies
                                                           First column: Frequency (Hz)
                                                           Second column: Magnitude (V/nT*Hz)
                                                           Third column: Phase  (deg)
                                                        - 'chopper_off' (ndarray, shape: n,3)
                                                           n - number of frequencies
                                                           First column: Frequency (Hz)
                                                           Second column: Magnitude (V/nT*Hz)
                                                           Third column: Phase  (deg)
                                          - 'hz' (dict) details of hx coil
                                                - 'sensor_type': 'MFS06e' (str)
                                                - 'sensor_serial_number': 301 (int)
                                                - 'chopper_status': 1 (int) 1 means 0N, 0 means OFF
                                                - 'calibration_data' (dict)
                                                    - '302' (dict) - key is coil number
                                                        - 'chopper_on' (ndarray, shape: n,3)
                                                           n - number of frequencies
                                                           First column: Frequency (Hz)
                                                           Second column: Magnitude (V/nT*Hz)
                                                           Third column: Phase  (deg)
                                                        - 'chopper_off' (ndarray, shape: n,3)
                                                           n - number of frequencies
                                                           First column: Frequency (Hz)
                                                           Second column: Magnitude (V/nT*Hz)
                                                           Third column: Phase  (deg)
        :type calibration_data_magnetic: dict
        :param apply_notch_filter: A boolean flag indicating whether a notch filter should be applied. Default is False.
        :type apply_notch_filter: bool
        :param notch_frequency: The frequency to be removed using the notch filter. Default is None.
        :type notch_frequency: float
        :param notch_harmonics: Defines the number of harmonics to be filtered using a notch filter.
                                If set to `None` (default), all harmonics up to the highest frequency of interest
                                will be removed. Else, if `notch_frequency` is 50 Hz and `notch_harmonics`
                                is 3, the harmonics at 100 Hz, 150 Hz and 200 Hz will be filtered out.
        :type notch_harmonics: Optional[int]
        :param process_mt: A boolean flag indicating whether the MT impedance is estimated during processing.
                           Default is True. Set to False to skip MT impedance calculations if only Tipper is
                           being estimated.
        :type process_mt: bool
        :param process_tipper: A boolean flag indicating whether the Tipper is estimated during processing.
                               Default is True. Set to False if Tipper is not part of the survey to avoid unnecessary
                               calculations.

        :return: None
        :rtype: NoneType

        """
        # Set attributes from parameters
        self.sampling_frequency = sampling_frequency
        self.fft_length = fft_length
        self.parzen_window_radius = parzen_window_radius
        self.overlap = overlap
        self.remote_reference = remote_reference
        self.calibration_data_electric = calibration_data_electric
        self.calibration_data_magnetic = calibration_data_magnetic
        self.notch_frequency = notch_frequency
        self.notch_harmonics = notch_harmonics
        self.process_mt = process_mt
        self.process_tipper = process_tipper

        # Class related attributes
        self.channels = None
        self.fft_frequencies = None
        self.band_averaged_dataset = None
        self.avgf = None
        self.dof = None
        self.spectra = None

        # Dividing time series into several time windows of length equals to fft length. Number of windows will
        # depend on the time series overlap.
        self.time_series = _reshape_time_series_with_overlap(time_series=time_series, fft_length=fft_length, overlap=overlap)
        del time_series

        self.ft_list = utils.get_target_frequency_list(sampling_frequency=self.sampling_frequency,
                                                       parzen_window_radius=self.parzen_window_radius,
                                                       fft_length=self.fft_length,
                                                       table_type='default',
                                                       frequencies_per_decade=frequencies_per_decade)

        self.get_channels()  # Get list of available ts channel. 'Ex', 'Ey', ....
        if calibrate_electric:
            self.calibrate_electric()
        if apply_notch_filter:
            self.apply_notch()
        self.detrend_time_series()
        self.perform_fft()
        if calibrate_magnetic:
            self.calibrate_magnetic()
        del self.calibration_data_magnetic
        del self.time_series
        gc.collect()
        self.perform_band_averaging()

    def get_channels(self) -> None:
        """
        Get details of channels in the received data.

        :return: None
        :rtype: NoneType

        """

        self.channels = list(self.time_series.keys())

    def calibrate_electric(self) -> None:
        """
        Calibrate electric field data

        :return: None
        :rtype: NoneType

        """
        print('Calibrating electric field channels.')
        if 'ex' in self.channels:
            dipole_ns = abs(self.calibration_data_electric['ex']['x1']) + abs(self.calibration_data_electric['ex']['x2'])
            self.time_series['ex'] = self.time_series['ex'] / (1 * dipole_ns / 1000)
        if 'ey' in self.channels:
            dipole_ew = abs(self.calibration_data_electric['ey']['y1']) + abs(self.calibration_data_electric['ey']['y2'])
            self.time_series['ey'] = self.time_series['ey'] / (1 * dipole_ew / 1000)

    def calibrate_magnetic(self) -> None:
        """
        Calibrates the magnetic field channels.

        :return: None
        :rtype: NoneType

        """
        print('Calibrating magnetic field channels.')
        desired_elements = ['hx', 'hy', 'hz', 'rx', 'ry']
        # Create a list mag channels out of desired elements if existing in self.channels
        magnetic_channels = [element for element in desired_elements if element in self.channels]
        for channel in magnetic_channels:
            if self.calibration_data_magnetic['instrument'] == 'metronix':
                sensor_type = self.calibration_data_magnetic[channel]['sensor_type']
                sensor_serial_number = str(self.calibration_data_magnetic[channel]['sensor_serial_number'])
                if self.calibration_data_magnetic[channel]['chopper_status'] == 1:
                    chopper_status = "chopper_on"
                elif self.calibration_data_magnetic[channel]['chopper_status'] == 0:
                    chopper_status = "chopper_off"
                else:
                    chopper_status = None
                calibration_data = self.calibration_data_magnetic[channel]['calibration_data'][sensor_serial_number][
                    chopper_status]
                calibration_object = MetronixCalibration(self.spectra[channel], self.fft_frequencies,
                                                         sensor_type,
                                                         chopper_status, calibration_data)
                self.spectra[channel] = calibration_object.calibrated_data
            else:
                raise NotImplementedError(
                    f'Calibration for {self.calibration_data_magnetic["instrument"]} instruments is not implemented as of now.'
                )

    def apply_notch(self) -> None:
        """
        Apply the notch filter

        :return: None
        :rtype: NoneType

        """
        print("Applying notch filter...")
        # TODO: Replace this with multiprocessing
        with ThreadPoolExecutor(max_workers=len(self.time_series)) as executor:
            # Dictionary to hold futures
            futures = {}

            # Submit each task to the executor
            for channel in self.channels:
                futures[channel] = executor.submit(sp.notch_filter_sos,
                                                   time_series=self.time_series[channel],
                                                   sampling_frequency=self.sampling_frequency,
                                                   notch_frequency=self.notch_frequency,
                                                   harmonics=self.notch_harmonics)

            # Retrieve the results when needed
            for channel, future in futures.items():
                self.time_series[channel] = future.result()
        print("Notch filter applied.")

    def detrend_time_series(self) -> None:
        """
        Detrend time series

        :return: None
        :rtype: NoneType

        """
        print('Applying detrend to the time series.')
        for channel in self.channels:
            self.time_series[channel] = signal.detrend(self.time_series[channel], axis=0)

    def perform_fft(self) -> None:
        """
        Perform FFT

        :return: None
        :rtype: NoneType

        """
        print('Performing FFT.')
        self.spectra = {}
        for channel in self.channels:
            self.fft_frequencies, self.spectra[channel] = sp.do_fft(self.time_series[channel], self.sampling_frequency,
                                                                    self.fft_length)

    def perform_band_averaging(self) -> None:
        """
        Perform band averaging.

        :return: None
        :rtype: NoneType

        """
        print('Starting band averaging.')

        # Creating empty arrays
        self.dof = np.empty(self.ft_list.shape[0], dtype=int)  # Degree of freedom
        self.avgf = np.empty(self.ft_list.shape[0], dtype=int)  # Number of frequencies used for averaging

        # Create a 3D array for parzen window for all target frequencies
        parzen_window = np.empty(
            (self.spectra[next(iter(self.spectra))].shape[0], 1, self.ft_list.shape[0]), dtype=float)

        # Populating parzen window arrays based on target frequency and window radius
        for i in range(self.ft_list.shape[0]):
            ft = float(self.ft_list[i])
            parzen_window[:, :, i] = stats.parzen(self.fft_frequencies, ft, self.parzen_window_radius)
            self.dof[i] = (2 * 2 * np.sum(parzen_window[:, :, i] != 0)) - 4
            self.avgf[i] = np.sum(parzen_window[:, :, i] != 0)

        # Create an empty xarray dataset
        self.band_averaged_dataset = xr.Dataset(
            coords={
                'time_window': np.arange(self.spectra[next(iter(self.spectra))].shape[1]),
                'frequency': self.ft_list
            }
        )

        # Precompute the sum of the Parzen window values to optimize performance
        # by avoiding redundant calculations during subsequent operations.
        sum_parzen = np.sum(parzen_window, axis=0)

        # =========== Band averaging ===============================

        # Calculate the cross-power spectra for Hx and Hy, which are common for both MT and Tipper.
        self.band_averaged_dataset['hx'] = (
            ('time_window', 'frequency'),
            np.sum(self.spectra['hx'][:, :, np.newaxis] * parzen_window, axis=0) / sum_parzen)
        self.band_averaged_dataset['hy'] = (
            ('time_window', 'frequency'),
            np.sum(self.spectra['hy'][:, :, np.newaxis] * parzen_window, axis=0) / sum_parzen)

        if self.remote_reference:
            # If remote referencing is requested, assign the conjugate term to the remote channels.
            self.band_averaged_dataset['rx'] = (
                ('time_window', 'frequency'),
                np.sum(self.spectra['rx'][:, :, np.newaxis] * parzen_window, axis=0) / sum_parzen)
            self.band_averaged_dataset['ry'] = (
                ('time_window', 'frequency'),
                np.sum(self.spectra['ry'][:, :, np.newaxis] * parzen_window, axis=0) / sum_parzen)
            hx_conj = np.conj(self.spectra['rx'])
            hy_conj = np.conj(self.spectra['ry'])
        else:
            # Else, local channels
            hx_conj = np.conj(self.spectra['hx'])
            hy_conj = np.conj(self.spectra['hy'])

        self.band_averaged_dataset['hxhx'] = (('time_window', 'frequency'), np.sum(
            self.spectra['hx'][:, :, np.newaxis] * hx_conj[:, :, np.newaxis] * parzen_window,
            axis=0) / sum_parzen)
        self.band_averaged_dataset['hyhy'] = (('time_window', 'frequency'), np.sum(
            self.spectra['hy'][:, :, np.newaxis] * hy_conj[:, :, np.newaxis] * parzen_window,
            axis=0) / sum_parzen)

        self.band_averaged_dataset['hxhy'] = (('time_window', 'frequency'), np.sum(
            self.spectra['hx'][:, :, np.newaxis] * hy_conj[:, :, np.newaxis] * parzen_window,
            axis=0) / sum_parzen)
        self.band_averaged_dataset['hyhx'] = (('time_window', 'frequency'), np.sum(
            self.spectra['hy'][:, :, np.newaxis] * hx_conj[:, :, np.newaxis] * parzen_window,
            axis=0) / sum_parzen)

        # Common for MT and Tipper
        denominator = ((self.band_averaged_dataset['hxhx'] * self.band_averaged_dataset['hyhy']) -
                       (self.band_averaged_dataset['hxhy'] * self.band_averaged_dataset['hyhx']))

        # TODO: This may not be created in this class
        self.band_averaged_dataset['alpha_h_selection'] = xr.DataArray(
            np.full(self.band_averaged_dataset['hx'].shape, True),
            coords=self.band_averaged_dataset.coords,
            dims=self.band_averaged_dataset.dims
        )

        # Compute if mt channels are requested
        if self.process_mt:
            self.band_averaged_dataset['ex'] = (
                ('time_window', 'frequency'),
                np.sum(self.spectra['ex'][:, :, np.newaxis] * parzen_window, axis=0) / sum_parzen)
            self.band_averaged_dataset['ey'] = (
                ('time_window', 'frequency'),
                np.sum(self.spectra['ey'][:, :, np.newaxis] * parzen_window, axis=0) / sum_parzen)

            # Compute the auto- and cross-spectra
            ex_conj = np.conj(self.spectra['ex'])
            ey_conj = np.conj(self.spectra['ey'])

            self.band_averaged_dataset['exex'] = (('time_window', 'frequency'), np.sum(
                self.spectra['ex'][:, :, np.newaxis] * ex_conj[:, :, np.newaxis] * parzen_window,
                axis=0) / sum_parzen)
            self.band_averaged_dataset['eyey'] = (('time_window', 'frequency'), np.sum(
                self.spectra['ey'][:, :, np.newaxis] * ey_conj[:, :, np.newaxis] * parzen_window,
                axis=0) / sum_parzen)
            self.band_averaged_dataset['exey'] = (('time_window', 'frequency'), np.sum(
                self.spectra['ex'][:, :, np.newaxis] * ey_conj[:, :, np.newaxis] * parzen_window,
                axis=0) / sum_parzen)

            self.band_averaged_dataset['exhx'] = (('time_window', 'frequency'), np.sum(
                self.spectra['ex'][:, :, np.newaxis] * hx_conj[:, :, np.newaxis] * parzen_window,
                axis=0) / sum_parzen)
            self.band_averaged_dataset['exhy'] = (('time_window', 'frequency'), np.sum(
                self.spectra['ex'][:, :, np.newaxis] * hy_conj[:, :, np.newaxis] * parzen_window,
                axis=0) / sum_parzen)

            self.band_averaged_dataset['eyhx'] = (('time_window', 'frequency'), np.sum(
                self.spectra['ey'][:, :, np.newaxis] * hx_conj[:, :, np.newaxis] * parzen_window,
                axis=0) / sum_parzen)
            self.band_averaged_dataset['eyhy'] = (('time_window', 'frequency'), np.sum(
                self.spectra['ey'][:, :, np.newaxis] * hy_conj[:, :, np.newaxis] * parzen_window,
                axis=0) / sum_parzen)

            # Computing the MT impedance for all time windows
            zxx_num = ((self.band_averaged_dataset['hyhy'] * self.band_averaged_dataset['exhx']) -
                       (self.band_averaged_dataset['hyhx'] * self.band_averaged_dataset['exhy']))
            zxy_num = ((self.band_averaged_dataset['hxhx'] * self.band_averaged_dataset['exhy']) -
                       (self.band_averaged_dataset['hxhy'] * self.band_averaged_dataset['exhx']))
            self.band_averaged_dataset['zxx'] = zxx_num / denominator
            self.band_averaged_dataset['zxy'] = zxy_num / denominator
            #
            zyx_num = ((self.band_averaged_dataset['hyhy'] * self.band_averaged_dataset['eyhx']) -
                       (self.band_averaged_dataset['hyhx'] * self.band_averaged_dataset['eyhy']))
            zyy_num = ((self.band_averaged_dataset['hxhx'] * self.band_averaged_dataset['eyhy']) -
                       (self.band_averaged_dataset['hxhy'] * self.band_averaged_dataset['eyhx']))
            self.band_averaged_dataset['zyx'] = zyx_num / denominator
            self.band_averaged_dataset['zyy'] = zyy_num / denominator

            # Preparing selection arrays, this may be used for data selection later.
            # TODO: This may not be created in this class
            self.band_averaged_dataset['ex_selection_coh'] = xr.DataArray(
                np.full(self.band_averaged_dataset['ex'].shape, True),
                coords=self.band_averaged_dataset.coords,
                dims=self.band_averaged_dataset.dims
            )
            self.band_averaged_dataset['ey_selection_coh'] = xr.DataArray(
                np.full(self.band_averaged_dataset['ey'].shape, True),
                coords=self.band_averaged_dataset.coords,
                dims=self.band_averaged_dataset.dims
            )

            # TODO: This may not be created in this class
            self.band_averaged_dataset['alpha_e_selection'] = xr.DataArray(
                np.full(self.band_averaged_dataset['ex'].shape, True),
                coords=self.band_averaged_dataset.coords,
                dims=self.band_averaged_dataset.dims
            )

        # Compute if tipper channels are requested
        if self.process_tipper:
            self.band_averaged_dataset['hz'] = (
                ('time_window', 'frequency'),
                np.sum(self.spectra['hz'][:, :, np.newaxis] * parzen_window, axis=0) / sum_parzen)

            hz_conj = np.conj(self.spectra['hz'])

            self.band_averaged_dataset['hzhz'] = (('time_window', 'frequency'), np.sum(
                self.spectra['hz'][:, :, np.newaxis] * hz_conj[:, :, np.newaxis] * parzen_window,
                axis=0) / sum_parzen)
            self.band_averaged_dataset['hzhx'] = (('time_window', 'frequency'), np.sum(
                self.spectra['hz'][:, :, np.newaxis] * hx_conj[:, :, np.newaxis] * parzen_window,
                axis=0) / sum_parzen)
            self.band_averaged_dataset['hzhy'] = (('time_window', 'frequency'), np.sum(
                self.spectra['hz'][:, :, np.newaxis] * hy_conj[:, :, np.newaxis] * parzen_window,
                axis=0) / sum_parzen)

            # Computing the Tipper transfer function for all time windows
            self.band_averaged_dataset['tzx'] = ((self.band_averaged_dataset['hzhx'] * self.band_averaged_dataset[
                'hyhy']) - (self.band_averaged_dataset['hzhy'] * self.band_averaged_dataset['hyhx'])) / denominator
            self.band_averaged_dataset['tzy'] = ((self.band_averaged_dataset['hzhy'] * self.band_averaged_dataset[
                'hxhx']) - (self.band_averaged_dataset['hzhx'] * self.band_averaged_dataset['hxhy'])) / denominator

            # TODO: This may not be created in this class
            self.band_averaged_dataset['hz_selection_coh'] = xr.DataArray(
                np.full(self.band_averaged_dataset['hz'].shape, True),
                coords=self.band_averaged_dataset.coords,
                dims=self.band_averaged_dataset.dims
            )

            print('Band averaging finished.')


def _reshape_time_series_with_overlap(time_series, fft_length, overlap):
    """
    Reshape array with overlap.

    :return: None
    :rtype: NoneType

    """
    for channel in time_series:
        time_series[channel] = utils.reshape_array_with_overlap(window_length=fft_length,
                                                                overlap=overlap,
                                                                data=time_series[channel])

    return time_series



"""
Module containing statistics functions
"""

import numpy as np
import xarray as xr


def parzen(f: np.ndarray, ft: float, cr: float) -> np.ndarray:
    """
    Function to create a parzen window

    :param f: Array of frequencies from FFT
    :type f: np.ndarray
    :param ft: Target frequency
    :type ft: float
    :param cr: Parzen Window Radius
    :type cr: float
    :return: Parzen window. Length: FFT_length/2. Array shape: (n,1) - 1D
    :rtype: np.ndarray

    """
    pf = np.empty((np.shape(f)[0],))
    pf[:] = np.nan
    fr = cr * ft
    pf[0] = 0
    for i in range(1, np.shape(f)[0]):
        cond = abs(ft - f[i])
        if cond == 0:
            pf[i] = 1
        elif (cond > 0) and (cond < fr):
            u = (np.pi * abs(cond)) / fr
            pf[i] = (np.sin(u) / u) ** 4
        elif (cond > fr) or (cond == fr):
            pf[i] = 0
    pf = pf.reshape(-1, 1)
    return pf


def jackknife(z: xr.DataArray) -> complex:
    """
    Function to compute jackknife mean

    :param z: Data array of impedance values
    :type z: xr.DataArray
    :return: Jackknife mean
    :rtype: complex

    """
    n = len(z)
    jackknife_means = np.zeros(n, dtype=complex)

    for i in range(n):
        jackknife_sample = np.delete(z, i)
        jackknife_means[i] = np.mean(jackknife_sample)

    jackknife_mean = np.mean(jackknife_means)

    return jackknife_mean


def fisher_critical(dof: int) -> float:
    """
    Calculate the critical value of the F-distribution for a given
    degrees of freedom.
    Dr. Manoj C. Nair helped with this computation.

    :param dof: Degree of freedom
    :type dof: int
    :return: The critical value of F-distribution
    :rtype: float

    """

    if dof < 6.0:
        # For small degrees of freedom, use a simple exponential approximation
        f_critical = 5 ** (dof + 1.0)
    else:
        # For larger degrees of freedom, use a more accurate approximation
        z = 1.6446
        h = 2.0 / (1.0 / (dof - 5.0) + 1.0 / 3.0)
        alpha = (z * z) / 6.0 - 0.5
        w_t = (z * np.sqrt(h + alpha) / h -
               ((1.0 / 3.0) - 1.0 / (dof - 5.0)) *
               (alpha + (5.0 / 6.0) - (2.0 / 3.0) / h))
        f_critical = np.exp(2 * w_t)

    return f_critical



"""
Utility functions for the metronix
"""

import datetime
import fnmatch
import glob
import os
import xml.etree.ElementTree as ET

import h5py
import numpy as np
import pandas as pd
import xarray as xr
from PyQt5.QtWidgets import QMessageBox

from sigmt.utils.metronix import cal_from_xml


def loadsites(project_dir: str) -> list:
    """
    It lists all folders in a time_series folder in the project, which satisfies two conditions.
    1. Folder should not be empty
    2. Atleast one meas sub folder.
    This is to ensure, it reads proper Metronix sites.

    :param project_dir: Path to the project directory.
    :type project_dir: str
    :return: List of sites
    :rtype: List[str]

    """
    ts_path = os.path.join(project_dir, 'time_series')
    sites = []
    for folder_name in os.listdir(ts_path):
        folder_path = os.path.join(ts_path, folder_name)
        # Check if it is a directory
        if os.path.isdir(folder_path):
            # Check if the directory is not empty
            if os.listdir(folder_path):
                # Check if there is at least one folder starting with 'meas'
                if any(fnmatch.fnmatch(subfolder, 'meas*') for subfolder in os.listdir(folder_path) if
                       os.path.isdir(os.path.join(folder_path, subfolder))):
                    sites.append(folder_name)
    return sites


def list_meas_folders(directory: str) -> list:
    """
    List all folders that start with 'meas' in the given directory.

    This function takes the path to a directory and returns a list of all
    subdirectories that start with 'meas'.

    :param directory: Path to the directory to search.
    :type directory: str
    :return: List of paths to subdirectories starting with 'meas'.
    :rtype: List[str]

    """
    # List to store the values
    meas_folders = []
    # Iterate through the items in the given directory
    for item in os.listdir(directory):
        item_path = os.path.join(directory, item)
        # Check if the item is a directory and starts with "meas"
        if os.path.isdir(item_path) and item.startswith("meas"):
            meas_folders.append(item)
    return meas_folders


def get_sampling_frequency_from_xml(meas_paths: list) -> tuple:
    """
    Get a list of meas paths, loops through the meas folders, reads xml file for sampling frequency
    and chopper status.
    Zips the values, converts to list and returns.

    :param meas_paths: List containing measurement paths.
    :type meas_paths: List[str]
    :return: tuple of lists containing chopper statuses.
    :rtype: Tuple of list[str]

    """
    sampfreq = []
    chopper_value = []
    for meas in meas_paths:
        try:
            allfiles = os.listdir(meas)
            xml_file = next((file for file in allfiles if file.endswith('.xml')), None)
            if xml_file is None:
                raise FileNotFoundError("No XML file in meas folder! Cannot proceed further.")
            xml_file_path = os.path.join(meas, xml_file)
            mytree = ET.parse(xml_file_path)
            root = mytree.getroot()
            sample_freq = root.findall(".//sample_freq")
            sample_freq = float(sample_freq[-1].text)
            sampfreq.append(sample_freq)
            hchopper_value = int(root.find(".//hchopper").text)
            if hchopper_value == 1:
                chopper_status = "chopper_on"
            elif hchopper_value == 0:
                chopper_status = "chopper_off"
            else:
                chopper_status = "unknown"
            chopper_value.append(chopper_status)
        except Exception as e:
            QMessageBox.critical(None, "Error", str(e))
    return sampfreq, chopper_value


def read_ts(meas_path: str, project_setup: dict) -> tuple:
    """
    Reads all time series files (ats) from the measurement path and returns header and time series data as dictionaries.

    :param meas_path: Measurement path
    :type meas_path: str
    :param project_setup: Dictionary containing project information
    :type project_setup: dict
    :return: Header information of time series and Time series data as dictionaries.
    :rtype: Tuple

    """
    header = {}
    ts = {}
    ats_files = glob.glob(os.path.join(meas_path, '*.ats'))
    if not ats_files:
        QMessageBox.information(None, "Information",
                                f"No time series files at {meas_path}. Skipping to next measurement.")
        return {}, {}
    for tspath in ats_files:
        [header_info, ts_data] = read_ats(tspath)
        # Gets start time from header
        start_time = datetime.datetime.utcfromtimestamp(header_info['start'][0])
        # Creates time index
        time_index = pd.date_range(start=start_time, periods=header_info['nsamples'],
                                   freq=str(int(1 / header_info['sfreq'][0] * 1e9)) + 'ns')
        duration = time_index.max() - time_index.min()
        header_info['duration'] = duration  # Saves duration just for information
        header[header_info['ch_type']] = header_info
        # Creates xarray
        if project_setup['processing_mode'] == 'MT + Tipper':
            if header_info['ch_type'] in ['ex', 'ey', 'hx', 'hy', 'hz']:
                ts[header_info['ch_type']] = xr.DataArray(ts_data, coords={'time': time_index}, dims=['time'])
        elif project_setup['processing_mode'] == 'MT Only':
            if header_info['ch_type'] in ['ex', 'ey', 'hx', 'hy']:
                ts[header_info['ch_type']] = xr.DataArray(ts_data, coords={'time': time_index}, dims=['time'])
    return header, ts


def read_ats(filename: str) -> tuple:
    """
    Reads Metronix time series (ATS) data file.
    This function reads both the header information and the time series data from a Metronix ATS file.

    :param filename: List containing measurement paths.
    :type filename: Str
    :return: Header information (Dict) from the ATS file and Time series data (ndarray[Float])
    :rtype: tuple
    """
    with open(filename, 'rb') as f:
        header = {}
        header['length'] = np.fromfile(
            f, dtype=np.int16, count=1).tolist()  # Header length
        header['ver'] = np.fromfile(
            f, dtype=np.int16, count=1).tolist()  # Header version
        header['nsamples'] = int(np.fromfile(f, dtype=np.int32, count=1)[0])  # Number of samples
        header['sfreq'] = np.fromfile(
            f, dtype=np.float32, count=1).tolist()  # sampling frequency, Hz
        # Start time, seconds since 1970
        header['start'] = np.fromfile(f, dtype=np.int32, count=1).tolist()
        header['lsb'] = np.fromfile(
            f, dtype=np.double, count=1).tolist()  # LSB-Value
        header['iGMTOffset'] = np.fromfile(f, dtype=np.int32, count=1).tolist()
        header['rOrigSampleFreq'] = np.fromfile(
            f, dtype=np.float32, count=1).tolist()
        header['adu_serial_number'] = np.fromfile(
            f, dtype=np.int16, count=1).tolist()  # ADU serial number
        header['adu_ADB'] = np.fromfile(
            f, dtype=np.int16, count=1).tolist()  # ADU serial number
        header['ch_no'] = np.fromfile(
            f, dtype=np.int8, count=1).tolist()  # Channel number
        header['bychopper'] = np.fromfile(
            f, dtype=np.int8, count=1).tolist()  # Chopper
        header['ch_type'] = np.fromfile(
            f, dtype=np.int8, count=2).tolist()  # channel type (Hx,Hy,...)
        header['ch_type'] = ''.join([chr(item) for item in header['ch_type']]).lower()
        header['sensor'] = np.fromfile(f, dtype=np.int8, count=6).tolist()
        header['sensor'] = ''.join([chr(item) for item in header['sensor']])
        header['sensor_no'] = np.fromfile(
            f, dtype=np.int16, count=1).tolist()  # Sensor serial number
        # x1 coordinate of 1. Dipole (m) - NS
        header['x1'] = np.fromfile(f, dtype=np.float32, count=1).tolist()
        # y1 coordinate of 1. Dipole (m) - EW
        header['y1'] = np.fromfile(f, dtype=np.float32, count=1).tolist()
        # z1 coordinate of 1. Dipole (m)
        header['z1'] = np.fromfile(f, dtype=np.float32, count=1).tolist()
        # x2 coordinate of 1. Dipole (m) - NS
        header['x2'] = np.fromfile(f, dtype=np.float32, count=1).tolist()
        # y2 coordinate of 1. Dipole (m) - EW
        header['y2'] = np.fromfile(f, dtype=np.float32, count=1).tolist()
        # z2 coordinate of 1. Dipole (m)
        header['z2'] = np.fromfile(f, dtype=np.float32, count=1).tolist()
        header['dipole_length'] = np.fromfile(
            f, dtype=np.float32, count=1).tolist()
        header['dipole_angle'] = np.fromfile(
            f, dtype=np.float32, count=1).tolist()
        header['rProbeRes'] = np.fromfile(
            f, dtype=np.float32, count=1).tolist()
        header['rDCOffset'] = np.fromfile(
            f, dtype=np.float32, count=1).tolist()
        header['rPreGain'] = np.fromfile(f, dtype=np.float32, count=1).tolist()
        header['rPostGain'] = np.fromfile(
            f, dtype=np.float32, count=1).tolist()
        header['lat'] = np.fromfile(f, dtype=np.int32, count=1).tolist()
        header['lon'] = np.fromfile(f, dtype=np.int32, count=1).tolist()
        header['elev'] = np.fromfile(f, dtype=np.int32, count=1).tolist()
        header['byLatLongType'] = np.fromfile(
            f, dtype=np.int8, count=1).tolist()
        header['byLatLongType'] = ''.join(
            [chr(item) for item in header['byLatLongType']])
        header['byAddCoordType'] = np.fromfile(
            f, dtype=np.int8, count=1).tolist()
        header['byAddCoordType'] = ''.join(
            [chr(item) for item in header['byAddCoordType']])
        header['siRefMedian'] = np.fromfile(
            f, dtype=np.int16, count=1).tolist()
        header['dblNorthing'] = np.fromfile(
            f, dtype=np.float64, count=1).tolist()
        header['dblEasting'] = np.fromfile(
            f, dtype=np.float64, count=1).tolist()
        header['byGPSStat'] = np.fromfile(f, dtype=np.int8, count=1).tolist()
        header['byGPSStat'] = ''.join([chr(item)
                                       for item in header['byGPSStat']])
        header['byGPSAccuracy'] = np.fromfile(
            f, dtype=np.int8, count=1).tolist()
        header['byGPSAccuracy'] = ''.join(
            [chr(item) for item in header['byGPSAccuracy']])
        header['iUTCOffset'] = np.fromfile(f, dtype=np.int16, count=1).tolist()
        header['achSystemType'] = np.fromfile(
            f, dtype=np.int8, count=12).tolist()
        header['achSystemType'] = ''.join(
            [chr(item) for item in header['achSystemType']])
        header['achSurveyHeaderName'] = np.fromfile(
            f, dtype=np.int8, count=12).tolist()
        header['achSurveyHeaderName'] = ''.join(
            [chr(item) for item in header['achSurveyHeaderName']])
        header['achMeasType'] = np.fromfile(f, dtype=np.int8, count=4).tolist()
        header['achMeasType'] = ''.join(
            [chr(item) for item in header['achMeasType']])
        header['DCOffsetCorrValue'] = np.fromfile(
            f, dtype=np.float64, count=1).tolist()
        header['DCOffsetCorrOn'] = np.fromfile(
            f, dtype=np.int8, count=1).tolist()
        header['InputDivOn'] = np.fromfile(f, dtype=np.int8, count=1).tolist()
        header['bit_indicator'] = np.fromfile(
            f, dtype=np.int16, count=1).tolist()
        header['achSelfTestResult'] = np.fromfile(
            f, dtype=np.int8, count=2).tolist()
        header['achSelfTestResult'] = ''.join(
            [chr(item) for item in header['achSelfTestResult']])
        header['numslice'] = np.fromfile(f, dtype=np.uint16, count=1).tolist()
        header['siCalFreqs'] = np.fromfile(
            f, dtype=np.uint16, count=1).tolist()
        f.seek(header['length'][0])
        ts = np.fromfile(f, dtype=np.int32, count=header['nsamples'])
        ts = ts * header['lsb'][0]
        return header, ts


def read_calibration_from_xml(meas_path: str) -> dict:
    """
    Reads calibration data from xml files.

    :param meas_path: Path to the measurement file, for eg: 'D:/project\\time_series\\KL33A\\meas_2019-08-29_14-30-00'
    :type meas_path: str
    :return: The Dictionary contains calibration data, ['coil_number'], within that 'ChoppOn' and 'ChoppOff'
    :rtype: dict

    """
    caldata = {}
    xmlfiles = []
    for file in os.listdir(meas_path):
        if file.endswith(".xml"):
            xmlfiles.append(file)
    xmlfile = os.path.join(meas_path, xmlfiles[0])

    channels_info = cal_from_xml.extract_channel_info(xmlfile)

    # Filter out channels with types 'Ex' and 'Ey'
    filtered_channels = cal_from_xml.filter_channels(channels_info, ['Ex', 'Ey'])
    filtered_channel_ids = [channel[0] for channel in filtered_channels]

    for filtered_channel_id, val in zip(filtered_channel_ids, filtered_channels):
        caldata[val[2]] = cal_from_xml.read_cal_data(xmlfile, filtered_channel_id)

    return caldata


def prepare_ts_from_h5(h5file_path: str, ts_key: str) -> dict:
    """
    Read time series from the h5 file.

    :param h5file_path: Path to the database file
    :type h5file_path: str
    :param ts_key: Key value for to access from the h5 group. Eg: 'ts_0'
    :type ts_key: str
    :return: Dictionary containing time series for 'Ex', 'Ey', ...
    :rtype: dict

    """
    ts = {}
    with h5py.File(h5file_path, 'r') as f:
        channel_keys = f[ts_key].keys()
        for channel in channel_keys:
            ts[channel] = f[ts_key][channel][:]

    return ts



"""
Class for the robust estimation of MT impedance and tipper
"""
import time

import numpy as np
import xarray as xr
from sklearn.covariance import MinCovDet

import sigmt.core.statistics as stats


class RobustEstimation:
    """
    Class for robust estimation
    """

    def __init__(self, procinfo: dict, dataset: xr.Dataset) -> None:
        """
        Constructor

        :param procinfo: Dictionary containing information for processing such as target
                        frequency, local and remote station details, FFT length, parzen radius,
                        location, etc.
                        :TODO make it instrument independent
        :type procinfo: dict
        :param dataset: xarray dataset containing required numpy arrays for the regression.
                        Cross powers, Selection booleans, TF estimates for all time windows.

        :return: None
        :rtype: NoneType

        """
        self.ft = None
        self.output_channels = None
        self.channel = None
        self.filtered_dataset = None
        self.estimates = None
        self.z2_initial_jackknife = None
        self.z1_initial_jackknife = None
        self.z2_mean_md = None
        self.z1_mean_md = None
        self.output = None
        self.z1_num_keys = None
        self.z2_num_keys = None
        self.z_deno_keys = None
        self.residuals = None
        self.huber_weights = None
        self.z1_robust_huber = None
        self.z2_robust_huber = None
        self.z1_var = None
        self.z2_var = None
        self.predicted_coherency = None
        self.procinfo = procinfo
        self.processing_mode = self.procinfo['processing_mode']
        self.get_output_channels()
        self.dataset = dataset
        self.target_frequencies = dataset['frequency'].values
        self.get_estimate_and_variance()

    def get_output_channels(self) -> None:
        """
        Creates output channels based on processing mode.
        SigMT can support MT+Tipper, MT Only and Tipper Only mode.

        :return: None
        :rtype: NoneType

        """
        if self.processing_mode == "MT + Tipper":
            self.output_channels = ['ex', 'ey', 'hz']
        elif self.processing_mode == "MT Only":
            self.output_channels = ['ex', 'ey']
        elif self.processing_mode == "Tipper Only":
            self.output_channels = ['hz']

    def get_estimate_and_variance(self) -> None:
        """
        Computes the robust estimates and variances.
        This method loops over available output channels and target frequencies.
        It rejects noisy windows and pass data to robust estimation and variance
        calculation methods and saves data in an xarray.

        :return: None
        :rtype: NoneType

        """
        estimate = {}
        estimation_time = time.time()
        for self.channel in self.output_channels:
            z1 = []
            z2 = []
            z1_var = []
            z2_var = []
            coh = []
            for self.ft in self.target_frequencies:
                single_time = time.time()
                # Filtering based on target frequency
                self.filtered_dataset = self.dataset.sel(frequency=self.ft)
                # Filtering based on a selection array for ex/ey/hz
                self.get_selection_array()
                selected_windows = \
                    np.where(self.filtered_dataset[f'selection_array_{self.channel}'].values.astype(
                        bool))[0]
                print(
                    f'Channel: {self.channel}, ft: {self.ft} Hz. Applying data selection conditions.')
                self.filtered_dataset = self.filtered_dataset.isel(time_window=selected_windows)
                # Filtering based on mahalanobis distance
                self.get_mahalanobis_distance()
                mahalanobis_selection = self.filtered_dataset['maha_dist'] < self.procinfo[
                    'md_thresh']
                selected_windows = np.where(mahalanobis_selection.astype(bool))[0]
                print(
                    f'Channel: {self.channel}, ft: {self.ft} Hz. Applying Mahalanobis distance condition.')
                self.filtered_dataset = self.filtered_dataset.isel(time_window=selected_windows)
                print(
                    f'Channel: {self.channel}, ft: {self.ft} Hz. Getting Jackknife initial guess.')
                self.get_jackknife_initial_guess()
                print(f'Channel: {self.channel}, ft: {self.ft} Hz. Performing robust estimation.')
                self.perform_robust_estimation()
                print(f'Channel: {self.channel}, ft: {self.ft} Hz. Computing variance.')
                self.get_variance()
                z1.append(self.z1_robust_huber)
                z2.append(self.z2_robust_huber)
                z1_var.append(self.z1_var)
                z2_var.append(self.z2_var)
                coh.append(self.predicted_coherency)
                print(f'Time taken: {time.time() - single_time} seconds.')
            #
            if self.channel == 'ex':
                z1_key = "zxx"
                z2_key = "zxy"
                z1_var_key = "zxx_var"
                z2_var_key = "zxy_var"
                coh_key = 'coh_ex'
            if self.channel == 'ey':
                z1_key = "zyx"
                z2_key = "zyy"
                z1_var_key = "zyx_var"
                z2_var_key = "zyy_var"
                coh_key = 'coh_ey'
            if self.channel == 'hz':
                z1_key = "tzx"
                z2_key = "tzy"
                z1_var_key = "tzx_var"
                z2_var_key = "tzy_var"
                coh_key = 'coh_hz'
            estimate[z1_key] = xr.DataArray(np.asarray(z1),
                                            coords={'frequency': self.target_frequencies},
                                            dims='frequency'
                                            )
            estimate[z2_key] = xr.DataArray(np.asarray(z2),
                                            coords={'frequency': self.target_frequencies},
                                            dims='frequency'
                                            )
            estimate[z1_var_key] = xr.DataArray(np.asarray(z1_var),
                                                coords={'frequency': self.target_frequencies},
                                                dims='frequency'
                                                )
            estimate[z2_var_key] = xr.DataArray(np.asarray(z2_var),
                                                coords={'frequency': self.target_frequencies},
                                                dims='frequency'
                                                )
            estimate[coh_key] = xr.DataArray(np.asarray(coh),
                                             coords={'frequency': self.target_frequencies},
                                             dims='frequency'
                                             )
            self.estimates = xr.Dataset(estimate)
        print(f'Total time taken for robust estimation: {time.time() - estimation_time} seconds.')

    def get_selection_array(self) -> None:
        """
        Prepare an array of boolean suggesting selected time windows.

        :return: None
        :rtype: NoneType

        """
        self.filtered_dataset['selection_array_ex'] = (
                self.filtered_dataset['ex_selection_coh'] * self.filtered_dataset['alpha_e_selection'] *
                self.filtered_dataset['alpha_h_selection'])
        self.filtered_dataset['selection_array_ey'] = (
                self.filtered_dataset['ey_selection_coh'] * self.filtered_dataset['alpha_e_selection'] *
                self.filtered_dataset['alpha_h_selection'])
        if not self.processing_mode == "MT Only":
            self.filtered_dataset['selection_array_hz'] = (
                    self.filtered_dataset['hz_selection_coh'] * self.filtered_dataset['alpha_e_selection'] *
                    self.filtered_dataset['alpha_h_selection'])
        # Avoiding GUI from crashing due to insufficient data
        if np.sum(self.filtered_dataset['selection_array_ex']) < 10:
            print(
                f'Skipping polarization direction selection for ft:{self.ft} Hz (ex) '
                f'due to insufficient data for robust regression. Try with different min and max values.')
            self.filtered_dataset['selection_array_ex'] = self.filtered_dataset['ex_selection_coh']
        #
        if np.sum(self.filtered_dataset['selection_array_ey']) < 10:
            print(
                f'Skipping polarization direction selection for ft:{self.ft} Hz (ey) '
                f'due to insufficient data for robust regression. Try with different min and max values.')
            self.filtered_dataset['selection_array_ey'] = self.filtered_dataset['ey_selection_coh']
        #
        if not self.processing_mode == "MT Only":
            if np.sum(self.filtered_dataset['selection_array_hz']) < 10:
                print(
                    f'Skipping polarization direction selection for ft:{self.ft} Hz (hz) '
                    f'due to insufficient data for robust regression. Try with different min and max values.')
                self.filtered_dataset['selection_array_hz'] = self.filtered_dataset[
                    'hz_selection_coh']

    def get_mahalanobis_distance(self) -> None:
        """
        Method to calculate the mahalanobis distance

        :return: None
        :rtype: NoneType

        """
        if self.channel == 'ex':
            z1 = self.filtered_dataset['zxx'].values
            z2 = self.filtered_dataset['zxy'].values
        if self.channel == 'ey':
            z1 = self.filtered_dataset['zyx'].values
            z2 = self.filtered_dataset['zyy'].values
        if self.channel == 'hz':
            z1 = self.filtered_dataset['tzx'].values
            z2 = self.filtered_dataset['tzy'].values
        data = np.transpose(np.vstack((z1.real, z1.imag, z2.real, z2.imag)))
        robust_cov = MinCovDet(random_state=0).fit(data)
        self.filtered_dataset['maha_dist'] = xr.DataArray(
            np.sqrt(robust_cov.mahalanobis(data)),
            coords=self.filtered_dataset.coords,
            dims=self.filtered_dataset.dims
        )
        self.z1_mean_md = complex(robust_cov.location_[0], robust_cov.location_[1])
        self.z2_mean_md = complex(robust_cov.location_[2], robust_cov.location_[3])

    def get_jackknife_initial_guess(self) -> None:
        """
        Prepares jackknife initial guess from transfer function values from selected time windows.

        :return: None
        :rtype: NoneType

        """
        if self.channel == 'ex':
            self.z1_initial_jackknife = stats.jackknife(self.filtered_dataset['zxx'])
            self.z2_initial_jackknife = stats.jackknife(self.filtered_dataset['zxy'])
        if self.channel == 'ey':
            self.z1_initial_jackknife = stats.jackknife(self.filtered_dataset['zyx'])
            self.z2_initial_jackknife = stats.jackknife(self.filtered_dataset['zyy'])
        if self.channel == 'hz':
            self.z1_initial_jackknife = stats.jackknife(self.filtered_dataset['tzx'])
            self.z2_initial_jackknife = stats.jackknife(self.filtered_dataset['tzy'])

    def perform_robust_estimation(self) -> None:
        """
        Performs robust estimation.

        The robust estimation algorithm used here is adapted from the following sources:

        1.  Ritter, O., Junge, A., Dawes, G.J., 1998. New equipment and processing for
            magnetotelluric remote reference observations. Geophys. J. Int. 132 (3), 535–548.
            https://doi.org/10.1046/j.1365-246X.1998.00440.x.

        2.  Manoj, C., 2003. Magnetotelluric Data Analysis Using Advances in Signal Processing
            Techniques. Ph.D. Dissertation. CSIR – National Geophysical Research Institute and
            Osmania University, India.

        3.  Chave, A., Jones, A. (Eds.), 2012. The Magnetotelluric Method: Theory and Practice.
            Cambridge University Press, Cambridge. Pages: 188-189

        :return: None
        :rtype: NoneType

        """
        self.z1_robust_huber = None
        self.z2_robust_huber = None
        prev_sum_squared = None

        self.get_keys()
        hx = self.filtered_dataset['hx']
        hy = self.filtered_dataset['hy']
        n_time_windows = self.output.shape[0]

        z1 = self.z1_initial_jackknife
        z2 = self.z2_initial_jackknife

        # ------------ Iteration starts here ------------------
        for iteration in range(20):
            # Calculating residuals
            self.residuals = abs(self.output - (z1 * hx) - (z2 * hy))

            if iteration == 0:
                # Iteration = 0 produces the preliminary estimates of the transfer function based on MAD.
                # Robust estimation of transfer function (A1.3 - Ritter (1998)) begins with iteration = 1.
                #
                # Initial guess of variance based on MAD scale estimate
                dm = 1.483 * np.median(abs(self.residuals - np.median(self.residuals)))
                # Upper limit to the MAD scale estimate
                scale_factor = 1.5
                km = scale_factor * dm
                # Allowing up to 5% of the data by adjusting scale_factor = 1.5.
                # Because, in some cases, high residual values prevent Huber weight
                # conditions from being met, causing the runtime error when Lc becomes zero.
                while int(np.sum(self.residuals <= km)) < int(np.ceil(len(self.residuals) * 0.05)):
                    scale_factor = scale_factor + 0.1
                    km = scale_factor * dm
                # Get huber weights based on km
                self.get_huber_weights(km)
            else:
                # Number of windows in which weight = 1
                lc = np.sum((self.huber_weights == 1) * 1)
                # Weighted sum of squared residuals
                sum_squared = (n_time_windows / (lc ** 2)) * (
                    np.sum(self.huber_weights * (self.residuals ** 2)))
                if prev_sum_squared is not None:
                    change = abs(prev_sum_squared - sum_squared) / prev_sum_squared
                    if change < 0.01:  # 1%
                        # Stop iteration when there is no change in sum squared more than 1%
                        # This logic is taken from
                        # Chave, A., Jones, A. (Eds.), 2012. The Magnetotelluric Method:
                        # Theory and Practice. Cambridge University Press, Cambridge.
                        # Page: 189
                        break
                prev_sum_squared = sum_squared
                # New variance of the robust solution
                dh = np.sqrt(sum_squared)
                # Upper limit
                kh = float(1.5 * dh)
                # Get huber weights based on kh
                self.get_huber_weights(kh)

            # Applying weights to band averaged cross-spectra
            element_dict = {}
            element_avg_dict = {}
            for i in range(4):
                element_dict[f'z1_num_{i}'] = self.filtered_dataset[
                                                  self.z1_num_keys[i]].values * self.huber_weights
                element_dict[f'z2_num_{i}'] = self.filtered_dataset[
                                                  self.z2_num_keys[i]].values * self.huber_weights
                element_dict[f'z_deno_{i}'] = self.filtered_dataset[
                                                  self.z_deno_keys[i]].values * self.huber_weights
            # Weighted mean of cross-spectra
            for i in range(4):
                element_avg_dict[f'z1_num_avg_{i}'] = np.sum(element_dict[f'z1_num_{i}']) / np.sum(
                    self.huber_weights)
                element_avg_dict[f'z2_num_avg_{i}'] = np.sum(element_dict[f'z2_num_{i}']) / np.sum(
                    self.huber_weights)
                element_avg_dict[f'z_deno_avg_{i}'] = np.sum(element_dict[f'z_deno_{i}']) / np.sum(
                    self.huber_weights)

            #
            z_deno = ((element_avg_dict['z_deno_avg_0'] * element_avg_dict['z_deno_avg_1']) -
                      (element_avg_dict['z_deno_avg_2'] * element_avg_dict['z_deno_avg_3']))
            z1 = (((element_avg_dict['z1_num_avg_0'] * element_avg_dict['z1_num_avg_1']) -
                   (element_avg_dict['z1_num_avg_2'] * element_avg_dict['z1_num_avg_3'])) /
                  z_deno)
            z2 = (((element_avg_dict['z2_num_avg_0'] * element_avg_dict['z2_num_avg_1']) -
                   (element_avg_dict['z2_num_avg_2'] * element_avg_dict['z2_num_avg_3'])) /
                  z_deno)

        if self.channel == 'ex' or self.channel == 'ey':
            # Metronix specific correction
            self.z1_robust_huber = z1 * -1
            self.z2_robust_huber = z2 * -1
        else:
            self.z1_robust_huber = z1
            self.z2_robust_huber = z2

    def get_keys(self) -> None:
        """
        Prepares keys based on the current channel of processing.

        :return: None
        :rtype: NoneType

        """
        self.z_deno_keys = ["hxhx", "hyhy", "hxhy", "hyhx"]
        if self.channel == 'ex':
            self.output = self.filtered_dataset['ex']
            self.z1_num_keys = ["hyhy", "exhx", "hyhx", "exhy"]  # Zxx
            self.z2_num_keys = ["hxhx", "exhy", "hxhy", "exhx"]  # Zxy
        if self.channel == 'ey':
            self.output = self.filtered_dataset['ey']
            self.z1_num_keys = ["hyhy", "eyhx", "hyhx", "eyhy"]  # Zyx
            self.z2_num_keys = ["hxhx", "eyhy", "hxhy", "eyhx"]  # Zyy
        if self.channel == 'hz':
            self.output = self.filtered_dataset['hz']
            self.z1_num_keys = ["hyhy", "hzhx", "hyhx", "hzhy"]  # Zyx
            self.z2_num_keys = ["hxhx", "hzhy", "hxhy", "hzhx"]  # Zyy

    def get_huber_weights(self, upper_limit: float) -> None:
        """
        Prepares huber weights based on the upper limit.
        """
        huber_matrix1 = (self.residuals <= upper_limit) * 1
        huber_matrix2 = (self.residuals > upper_limit) * 1
        huber_matrix2 = huber_matrix2 * (upper_limit / self.residuals)
        self.huber_weights = (huber_matrix1 + huber_matrix2).values

    def get_variance(self) -> None:
        """
        Parametric estimation of variances. Based on the predicted coherency and degree of freedom.
        Dr. Manoj C. Nair helped with this computation.

        :return: None
        :rtype: NoneType

        """
        if self.channel == 'ex':
            outout = np.sum(self.filtered_dataset['exex'].values * self.huber_weights) / np.sum(
                self.huber_weights)
            outhx = np.sum(self.filtered_dataset['exhx'].values * self.huber_weights) / np.sum(
                self.huber_weights)
            outhy = np.sum(self.filtered_dataset['exhy'].values * self.huber_weights) / np.sum(
                self.huber_weights)
        if self.channel == 'ey':
            outout = np.sum(self.filtered_dataset['eyey'].values * self.huber_weights) / np.sum(
                self.huber_weights)
            outhx = np.sum(self.filtered_dataset['eyhx'].values * self.huber_weights) / np.sum(
                self.huber_weights)
            outhy = np.sum(self.filtered_dataset['eyhy'].values * self.huber_weights) / np.sum(
                self.huber_weights)
        if self.channel == 'hz':
            outout = np.sum(self.filtered_dataset['hzhz'].values * self.huber_weights) / np.sum(
                self.huber_weights)
            outhx = np.sum(self.filtered_dataset['hzhx'].values * self.huber_weights) / np.sum(
                self.huber_weights)
            outhy = np.sum(self.filtered_dataset['hzhy'].values * self.huber_weights) / np.sum(
                self.huber_weights)

        hxhx = np.sum(self.filtered_dataset['hxhx'].values * self.huber_weights) / np.sum(
            self.huber_weights)
        hxhy = np.sum(self.filtered_dataset['hxhy'].values * self.huber_weights) / np.sum(
            self.huber_weights)
        hyhx = np.sum(self.filtered_dataset['hyhx'].values * self.huber_weights) / np.sum(
            self.huber_weights)
        hyhy = np.sum(self.filtered_dataset['hyhy'].values * self.huber_weights) / np.sum(
            self.huber_weights)

        zpz = self.z1_robust_huber * np.conj(outhx) + self.z2_robust_huber * np.conj(outhy)
        zpx = self.z1_robust_huber * hxhx + self.z2_robust_huber * hyhx
        zpy = self.z1_robust_huber * hxhy + self.z2_robust_huber * hyhy
        zpzp = self.z1_robust_huber * np.conj(zpx) + self.z2_robust_huber * np.conj(zpy)
        zpzp = zpzp * outout

        coh = complex(np.where(np.abs(zpzp) > 0, zpz / np.sqrt(zpzp), 1 + 1j))
        coh = np.abs(coh)

        coh = float(np.where(coh > 1.0, 1 / coh, coh))

        coh = coh ** 2

        dof = int(self.filtered_dataset['dof'].values)

        fis = stats.fisher_critical(dof)

        d = (4 / dof) * fis * (1.0 - coh) * outout

        if abs(hxhx) * abs(hyhy) > 0:
            r2xy = (hxhy * hyhx) / (hxhx * hyhy)
        else:
            r2xy = 100

        if r2xy > 0.999:
            r2xy = 0.999

        if abs(hxhx * (1 - r2xy)) > 0:
            da = d / (hxhx * (1 - r2xy))
        else:
            da = 100

        if abs(hyhy * (1 - r2xy)) > 0:
            db = d / (hyhy * (1 - r2xy))
        else:
            db = 100

        self.z1_var = np.sqrt(abs(np.real(da)))
        self.z2_var = np.sqrt(abs(np.real(db)))
        #
        self.predicted_coherency = np.sqrt(coh)



"""
Dialog box with information about the package
"""
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QDialog, QVBoxLayout, QLabel

from sigmt.__version__ import __version__


class AboutDialog(QDialog):
    """
    Dialog box with information about the package
    """

    def __init__(self):
        super().__init__()

        # Set the title and size of the dialog
        self.citation_label = None

        self.get_window_details()

        # Create a layout for the dialog
        layout = QVBoxLayout()

        self.version_label = QLabel(f'SigMT version: {__version__}')
        self.version_label.setAlignment(Qt.AlignCenter)

        self.get_citation_label()
        self.citation_label.setOpenExternalLinks(True)  # Enable clickable link
        self.citation_label.setAlignment(Qt.AlignCenter)

        # Add the label to the layout
        layout.addWidget(self.version_label)
        layout.addWidget(self.citation_label)

        # Set the layout for the dialog
        self.setLayout(layout)

    def get_window_details(self) -> None:
        """
        Getting window details
        """
        self.setWindowTitle("About")
        self.setFixedSize(400, 200)
        self.setWindowIcon(QIcon(r'sigmt\images\sigmt.ico'))

    def get_citation_label(self) -> None:
        """
        Create citation label
        """
        # Add the citation label with a clickable DOI link using rich text (HTML)
        self.citation_label = QLabel('''
                                     If you are using SigMT, please cite the following paper:<br><br>
                                     Ajithabh, K. S., Patro, P. K. (2023). SigMT: An open-source 
                                     Python package for magnetotelluric data processing. <br>
                                     <i>Computers & Geosciences</i>, 171, 105270.<br>
                                     DOI: <a href="https://doi.org/10.1016/j.cageo.2022.105270">
                                     https://doi.org/10.1016/j.cageo.2022.105270</a>
                                     ''')



"""
Design of the GUI for project creation
"""
from typing import Any, Dict

from PyQt5.QtWidgets import QLabel, QDialog, QLineEdit, QVBoxLayout, QDialogButtonBox, QComboBox


class ProjectSetupDialog(QDialog):
    """
    Design of the dialog box that appears for project creation. User can enter details in it.
    """

    def __init__(self, parent=None, interface=None):
        """
        Constructor

        :param parent: Parent
        :type parent: object
        :param interface: Tells which instrument specific interface is used. Eg: 'Metronix'
        :type interface: str
        """
        super().__init__(parent)
        self.interface = interface
        self.setWindowTitle("Project Setup")
        layout = QVBoxLayout()

        self.project_name = QLineEdit('XYZ Project')
        self.country = QLineEdit('India')
        self.acquired_by_company = QLineEdit('XYZ Organization')
        #
        self.processing_mode = QComboBox()
        self.processing_mode.addItem('MT + Tipper')
        self.processing_mode.addItem('MT Only')
        # TODO: Support use of tipper only in future
        # self.processing_mode.addItem('Tipper Only')
        #
        self.frequencies_per_decade = QLineEdit('12')
        self.notch_frequency = QLineEdit('50')
        self.preferred_cal_file = QComboBox()
        self.preferred_cal_file.addItem('xml')
        # TODO: Support use of metronix text file in future
        # self.preferred_cal_file.addItem('metronix txt')

        layout.addWidget(QLabel("Project Name:"))
        layout.addWidget(self.project_name)

        layout.addWidget(QLabel("Country:"))
        layout.addWidget(self.country)

        layout.addWidget(QLabel("Acquired by:"))
        layout.addWidget(self.acquired_by_company)

        layout.addWidget(QLabel("Processing mode:"))
        layout.addWidget(self.processing_mode)

        layout.addWidget(QLabel("Frequencies per Decade:"))
        layout.addWidget(self.frequencies_per_decade)

        layout.addWidget(QLabel("Notch Frequency:"))
        layout.addWidget(self.notch_frequency)

        layout.addWidget(QLabel("Preferred calibration file:"))
        layout.addWidget(self.preferred_cal_file)

        buttons = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttons.accepted.connect(self.accept)
        buttons.rejected.connect(self.reject)
        layout.addWidget(buttons)

        self.setLayout(layout)

    def get_values(self) -> Dict[str, Any]:
        """
        Return the values entered in the dialog box by the user.

        :return: Dictionary of content from the dialog box
        :rtype: dict
        """
        return {
            "project_name": self.project_name.text(),
            "country": self.country.text(),
            "acquired_by": self.acquired_by_company.text(),
            "processing_mode": self.processing_mode.currentText(),
            "frequencies_per_decade": self.frequencies_per_decade.text(),
            "notch_frequency": self.notch_frequency.text(),
            "interface": self.interface,
            "preferred_cal_file": self.preferred_cal_file.currentText()
        }



from setuptools import setup

from sigmt.__version__ import __version__

with open('requirements.txt') as f:
    requirements = f.read().splitlines()

setup(
    name='sigmt',
    version=__version__,
    packages=['sigmt', 'sigmt.core',
              'sigmt.gui', 'sigmt.utils',
              'sigmt.utils.edi', 'sigmt.utils.metronix',
              'sigmt.gui.project_related'],
    include_package_data=True,
    install_requires=requirements,
    entry_points={
        'console_scripts': [
            'sigmt=sigmt.main:main',
        ],
    },
    python_requires='>=3.8',
    # Metadata
    author='Ajithabh K.S.',
    author_email='ajithabhks@gmail.com',
    description='A tool for magnetotelluric time series processing',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/ajithabhks/sigmt',
    license='GPL',
    classifiers=[
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Programming Language :: Python :: 3',
        'Development Status :: 4 - Beta',
    ],
)



"""
This is the first landing window for the SigMT
"""

import sys

from PyQt5.QtCore import Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import QApplication, QPushButton, QDialog, QVBoxLayout, QLabel

from sigmt.gui.metronix import MainWindow as Metronix

if hasattr(Qt, 'AA_EnableHighDpiScaling'):
    QApplication.setAttribute(Qt.AA_EnableHighDpiScaling, True)

if hasattr(Qt, 'AA_UseHighDpiPixmaps'):
    QApplication.setAttribute(Qt.AA_UseHighDpiPixmaps, True)


class Welcome(QDialog):
    """
    Design of the welcome page. User can select the instrument based processing.
    """

    def __init__(self) -> None:
        """
        Constructor
        """
        super().__init__()
        self.metronix_button = None
        self.main_window = None
        self.init_ui()

    def init_ui(self) -> None:
        """
        Initial interface design

        :return: None
        """
        self.setWindowTitle('Welcome to SigMT | A Tool for Magnetotelluric Data Processing')
        self.setWindowIcon(QIcon(r'sigmt\images\sigmt.ico'))
        self.setGeometry(100, 100, 500, 500)
        layout = QVBoxLayout()

        layout.addStretch()
        # Add a welcome message label
        welcome_label = QLabel("Welcome to <b>SigMT</b>!", self)
        layout.addWidget(welcome_label, alignment=Qt.AlignHCenter)

        layout.addSpacing(50)

        select_label = QLabel("Select Instrument:", self)
        layout.addWidget(select_label, alignment=Qt.AlignHCenter)

        self.metronix_button = QPushButton('Metronix', self)
        layout.addWidget(self.metronix_button, alignment=Qt.AlignHCenter)

        layout.addStretch()

        self.setLayout(layout)

        self.metronix_button.clicked.connect(self.open_metronix_mainwindow)

    def open_metronix_mainwindow(self) -> None:
        """
        To open the metronix specific main window and close welcome page

        :return: None
        """
        self.main_window = Metronix()
        self.main_window.show()
        self.close()


def main() -> None:
    """
    Main Function

    :return: None
    """
    app = QApplication(sys.argv)
    _ = app.font()
    entry = Welcome()
    entry.show()
    sys.exit(app.exec_())


if __name__ == '__main__':
    main()



"""
Functions to read calibration data from the xml files.
"""

import xml.etree.ElementTree as ET

import numpy as np


def extract_channel_info(xmlfile: str) -> list:
    """
    Extracts number of channels in the xml file. It gives the channel IDs. It is required to identify
    coil if some calibration sections are missing from xml.

    :param xmlfile: Path to the xml file
    :type xmlfile: str

    :return: Contains a list of tuples. Eg tuple: ('7', 'Hx', '300'). 7: channel id, 300: sensor number.
    :rtype: list
    """
    tree = ET.parse(xmlfile)
    root = tree.getroot()

    channels_info = []

    for configuration in root.iter('configuration'):
        # Iterate over each <channel> within the <configuration>
        for channel in configuration.findall('channel'):
            channel_id = channel.get('id')
            channel_type = channel.find('channel_type').text if channel.find('channel_type') is not None else None
            sensor_sernum = channel.find('sensor_sernum').text if channel.find('sensor_sernum') is not None else None
            channels_info.append((channel_id, channel_type, sensor_sernum))

    return channels_info


def filter_channels(data: list, excluded_types: list) -> list:
    """
    It is to exclude some channels which are not required. For eg: 'Ex', 'Ey'

    :param data: It is the list of the tuples. Eg: ('5', 'Ex', '0')
    :type data: list
    :param excluded_types: List of strings, to be excluded. Eg: 'Ex', 'Ey'
    :type excluded_types: list

    :return: List of channels (data above) after exclusion.
    :rtype: list
    """
    return [channel for channel in data if channel[1] not in excluded_types]


def read_cal_data(xmlfile: str, channel_id: str) -> dict:
    """
    Reads freq, amplitude, phase for chopper on and off, for a coil in xml file.

    :param xmlfile: Path to the xml file
    :type xmlfile: str
    :param channel_id: Channel id
    :type channel_id: str

    :return: Dictionary containing 'chopper_on' and 'chopper_off' data.
    :rtype: dict
    """
    mytree = ET.parse(xmlfile)
    root = mytree.getroot()
    #

    calsensors = root[2]

    calsensors.findall(f".//channel[@id='{channel_id}']//caldata[@chopper='on']/c1")

    caldata = {}
    freq = []
    for x in calsensors.findall(f".//channel[@id='{channel_id}']//caldata[@chopper='on']/c1"):
        freq.append(float(x.text))

    mag = []
    for x in calsensors.findall(f".//channel[@id='{channel_id}']//caldata[@chopper='on']/c2"):
        mag.append(float(x.text))

    phase = []
    for x in calsensors.findall(f".//channel[@id='{channel_id}']//caldata[@chopper='on']/c3"):
        phase.append(float(x.text))

    freq = np.asarray(freq).reshape(-1, 1)
    mag = np.asarray(mag).reshape(-1, 1)
    phase = np.asarray(phase).reshape(-1, 1)
    caldata['chopper_on'] = np.concatenate((freq, mag, phase), axis=1)

    freq = []
    for x in calsensors.findall(f".//channel[@id='{channel_id}']//caldata[@chopper='off']/c1"):
        freq.append(float(x.text))

    mag = []
    for x in calsensors.findall(f".//channel[@id='{channel_id}']//caldata[@chopper='off']/c2"):
        mag.append(float(x.text))

    phase = []
    for x in calsensors.findall(f".//channel[@id='{channel_id}']//caldata[@chopper='off']/c3"):
        phase.append(float(x.text))

    freq = np.asarray(freq).reshape(-1, 1)
    mag = np.asarray(mag).reshape(-1, 1)
    phase = np.asarray(phase).reshape(-1, 1)
    caldata['chopper_off'] = np.concatenate((freq, mag, phase), axis=1)

    return caldata
